/*
    icecastpipe.cpp  -  Icecast/Shoutcast forwarder

    Copyright (c) 2004 by Stefan Gehn <metz AT gehn.net>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include "icecastpipe.h"
#include "floodcastconfig.h"
#include "mp3stream.h"

#include <qtimer.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kdebug.h>


// average resolution a timer can work at, below this value timed sending is
// probably going wrong
// i.e. either too early or too late
// TODO: might be nice to adapt this to system-load so we can send more precise
#define DEFAULT_TIMER_LATENCY 20

// Maximum no. of shout_send calls allowed to fail while streaming, in
// case of a buggy connection this might eventually help
#define MAX_SEND_ERRORS 3


IcecastPipe::IcecastPipe(QObject *parent, const char * name)
	: QObject(parent, name),
	mShout(0),
	mStream(0)
{
	shout_init();
	mSendErrors = 0;
	mFirst = true;
	mSendTimer = new QTimer(this, "mSendTimer");
	connect(mSendTimer, SIGNAL(timeout()), this, SLOT(slotTransmitData()));
}


IcecastPipe::~IcecastPipe()
{
	stop();
	shout_shutdown();
}


void IcecastPipe::slotNewData(char *data, int length)
{
	unsigned int pos = mBuffer.size();
	mBuffer.resize(mBuffer.size()+length, QGArray::SpeedOptim);
	//concatenate the new data onto the buffer
	for(int i=0; i<length; i++)
	{
		mBuffer[pos+i] = data[i];
	}

	if (mFirst)
	{
		if (mBuffer.size() < 2048) // just started and buffer is not filled yet
		{
			return;
		}
		else
		{
			kdDebug(66666) << k_funcinfo <<
				"First buffer now filled, mBuffer size = " << mBuffer.size() << endl;
			mFirst = false;
		}
	}


	// are we waiting to send out new data?
	if (!mSendTimer->isActive())
	{
		int delay = shout_delay(mShout);
		//kdDebug(66666) << k_funcinfo << "delay = " << delay << endl;
		// it does not make sense to wait for less than 5msec because
		// usually system latency is way higher than that.
		// Furthermore delay can be negative, i.e. we whould send faster!
		if (delay > DEFAULT_TIMER_LATENCY)
		{
			mSendTimer->start(delay - DEFAULT_TIMER_LATENCY, true); // send data after the wanted delay
		}
		else
		{
			slotTransmitData();
		}
	}
	/*else
	{
		kdDebug(66666) << k_funcinfo << "NOT ALLOWED TO SEND NOW!, mBuffer size :" << mBuffer.size() << endl;
	}*/
}


void IcecastPipe::slotTransmitData()
{
	//kdDebug(66666) << k_funcinfo << "Sending with mBuffer size :" << mBuffer.size() << endl;
	int res = shout_send(mShout, (const unsigned char *)mBuffer.data(), mBuffer.size());
	if (res != SHOUTERR_SUCCESS)
	{
		mSendErrors++;
		if (mSendErrors >= MAX_SEND_ERRORS)
		{
			kdDebug(66666) << k_funcinfo << "shout_send failed: " << res << endl;
			// Immediately disconnect so we do not try to send again
			disconnect(mStream, SIGNAL(newData(char *, int)),
				this, SLOT(slotNewData(char *, int)));
			reconnectOrError(res);
		}
	}
	else
	{
		mBuffer.resize(0);
	}
}


void IcecastPipe::start()
{
	kdDebug(66666) << k_funcinfo << endl;
	int res;

	mFirst = true;
	mSendErrors = 0;
	mBuffer.resize(0);

	if(mShout || mStream)
	{
		kdDebug(66666) << k_funcinfo <<
			"INTERNAL ERROR: Shoutcast object or stream object already present, aborting!" << endl;
		return;
	}

	mShout = shout_new();
	if(!mShout)
	{
		KMessageBox::sorry(0,
			i18n("Floodcast was unable to initialize the icecast library." \
				" Please check your libshout installation."));
		return;
	}

	applyOptions();

	res = shout_open(mShout);
	if (res != SHOUTERR_SUCCESS)
	{
		reconnectOrError(res);
		return;
	}

	if(!mTitle.isEmpty())
	{
		//kdDebug(66666) << k_funcinfo << "Setting initial title" << endl;
		if (!sendTitle())
		{
			kdDebug(66666) << k_funcinfo << "Could not set initial title!" << endl;
			//stop(); // not necessarily a reason to disconnect again, streaming still might work
			return;
		}
	}

	// Create the MP3 stream
	kdDebug(66666) << k_funcinfo << "Creating new MP3Stream" << endl;
	mStream = new MP3Stream(this, "mp3stream");
	// This signal is called when the encoder has produced new data
	connect(mStream, SIGNAL(newData(char *, int)),
		this, SLOT(slotNewData(char *, int)));
	connect(mStream, SIGNAL(encoderDied()), this, SLOT(slotEncoderDied()));

	emit(statusChange(true));
}


void IcecastPipe::stop()
{
	kdDebug(66666) << k_funcinfo << "Called" << endl;

	mSendTimer->stop();
	mBuffer.resize(0);
	mFirst = false;

	delete mStream;
	mStream = 0;

	if(mShout)
	{
		kdDebug(66666) << k_funcinfo << "Removing Shoutcast object" << endl;
		shout_close(mShout);
		shout_free(mShout);
		mShout = 0;
	}
	emit(statusChange(false));
}


void IcecastPipe::slotEncoderDied()
{
	kdDebug(66666) << k_funcinfo << endl;
	stop();
}


void IcecastPipe::restart(int pauseSecs)
{
	kdDebug(66666) << k_funcinfo << "pauseSecs = " << pauseSecs << endl;
	stop();
	QTimer::singleShot(pauseSecs * 1000, this, SLOT(start()));
}


void IcecastPipe::applyOptions()
{
	int res;

	if (!mShout)
	{
		kdDebug(66666) << k_funcinfo <<
			"No Shoutcast object to apply options to!" << endl;
		return;
	}

	res = shout_set_host(mShout, FloodcastConfig::forwardHost());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	res = shout_set_port(mShout, FloodcastConfig::forwardPort());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	//shout_set_user(mShout, "source");

	res = shout_set_password(mShout, FloodcastConfig::forwardPassword());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	int proto = SHOUT_PROTOCOL_HTTP;
	switch(FloodcastConfig::forwardType())
	{
		case (FloodcastConfig::Shoutcast):
			proto = SHOUT_PROTOCOL_ICY;
			break;

		case (FloodcastConfig::Icecast1):
			proto = SHOUT_PROTOCOL_XAUDIOCAST;
			break;

		case (FloodcastConfig::Icecast2):
			proto = SHOUT_PROTOCOL_HTTP;
			break;
	}

	shout_set_protocol(mShout, proto);
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	res = shout_set_format(mShout, SHOUT_FORMAT_MP3);
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	shout_set_audio_info(mShout, SHOUT_AI_BITRATE,
		QString::number(FloodcastConfig::bitrate()));
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	shout_set_audio_info(mShout, SHOUT_AI_CHANNELS,
		FloodcastConfig::mono() ? "1" : "2");
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	shout_set_name(mShout, FloodcastConfig::streamTitle());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	shout_set_genre(mShout, FloodcastConfig::streamGenre());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	shout_set_url(mShout, FloodcastConfig::streamURL());
	if (res != SHOUTERR_SUCCESS)
		shoutcastError(res);

	//int shout_set_description(mShout, descr);
}


void IcecastPipe::setTitle(const QString &songtitle)
{
	if (songtitle != mTitle)
	{
		mTitle = songtitle;
		sendTitle();
	}
}


bool IcecastPipe::sendTitle()
{
	if (mShout == 0)
	{
		kdDebug(66666) << k_funcinfo << "Called with uninitialized libshout" << endl;
		return false;
	}

	shout_metadata_t *metadata = shout_metadata_new();
	if (!metadata)
		return false;
	shout_metadata_add(metadata, "song", mTitle.latin1());

	int res = shout_set_metadata(mShout, metadata);
	if (res != SHOUTERR_SUCCESS)
	{
		shoutcastError(res);
		return false;
	}

	shout_metadata_free(metadata);
	return true;
}


void IcecastPipe::shoutcastError(int err)
{
	QString msg;
	switch (err)
	{
		// Bad parameters, either nonsense or not applicable due to the current state of the connection.
		case SHOUTERR_INSANE:
			kdDebug(66666) << kdBacktrace(4) <<
				"Insane parameters passed to libshout function." << endl;
			break;

		//Indicates the function could not allocate the memory it required.
		case SHOUTERR_MALLOC:
			kdDebug(66666) << kdBacktrace(4) <<
				"Could not allocate memory." << endl;
			break;

		// Indicates a connection with the server could not be established.
		case SHOUTERR_NOCONNECT:
			msg = i18n("Could not connect to streaming-server");
			break;

		//Indicates the server refused to accept a login attempt.
		// This could be caused by a bad user name or password.
		case SHOUTERR_NOLOGIN:
			msg = i18n("The streaming-server refused your login. Username" \
				" or password might be wrong.");
			break;

		case SHOUTERR_SOCKET: //Indicates an error sending or receiving data.
			/*kdDebug(66666) << kdBacktrace(4) <<
				"Error sending or receiving socket data." << endl;*/
			msg = i18n("streaming connection was interrupted unexpectedly.");
			break;

		case SHOUTERR_METADATA:
			msg = i18n("Could not update stream-metadata on server");
			break;

		case SHOUTERR_CONNECTED:
			kdDebug(66666) << kdBacktrace(4) <<
				"While connected, you attempted to call a function which only " \
				"makes sense before connection (eg you attempted to set the user" \
				" name or stream name)." << endl;
			break;

		case SHOUTERR_UNCONNECTED:
			kdDebug(66666) << kdBacktrace(4) <<
				"Attempted to use a function that requires an open connection " \
				"while not being connected." << endl;
			break;

		case SHOUTERR_UNSUPPORTED:
			kdDebug(66666) << kdBacktrace(4) <<
				"Attempted to use a function which is unsupported in the state of" \
				" this connection." << endl;
			break;
	}

	if(!msg.isEmpty())
	{
		KMessageBox::queuedMessageBox(0, KMessageBox::Error, msg,
			i18n("Floodcast Error"));
	}
}

void IcecastPipe::reconnectOrError(int res)
{
	if (FloodcastConfig::forwardAutoreconnect())
	{
		kdDebug(66666) << k_funcinfo << "autoreconnect after " <<
			FloodcastConfig::forwardAutoreconnectTime() << " seconds" << endl;
		restart(FloodcastConfig::forwardAutoreconnectTime());
	}
	else
	{
		stop();
		shoutcastError(res);
	}
}


#include "icecastpipe.moc"
