/*
    mp3stream.cpp  -  Encodes raw artsd pcm data to mp3

    Copyright (c) 2002 by Ryan Cumming
    Copyright (c) 2004 by Stefan Gehn <metz AT gehn.net>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include <qtimer.h>
#include <kdebug.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kprocess.h>

#include <noatun/app.h>
#include <noatun/engine.h>
#include <noatun/player.h>
#include <noatun/noatunarts.h>

#include "floodcastconfig.h"
#include "mp3stream.h"

#define DATACHECK_TIME 120

MP3Stream::MP3Stream(QObject *parent, const char *name)
	: QObject(parent, name),
	mRawStream(0),
	_mId(0),
	_lame(0),
	_free_to_write(false),
	_audioTimer(0)
{
	kdDebug(66666) << k_funcinfo << endl;
	QTimer::singleShot(0, this, SLOT(start()));
}


MP3Stream::~MP3Stream()
{
	kdDebug(66666) << k_funcinfo << endl;
	stop();
}


void MP3Stream::timeout()
{
	if (mRawStream->dataWaiting())
	{
		//kdDebug(66666) << k_funcinfo << "data to send" << endl;
		if (_free_to_write)
		{
			std::vector<long> *data(mRawStream->queuedData());

			_lame->writeStdin((const char*)(&data->front()), sizeof(long) * data->size());
			_free_to_write = false;

			delete data;
		}
	}
	else
	{
		//kdDebug(66666) << k_funcinfo << "NO DATA TO SEND" << endl;
		// There was no data waiting for us, sleep
		_audioTimer->start(DATACHECK_TIME, true);
	}
}


void MP3Stream::receivedStdout(KProcess */*proc*/, char *buffer, int buflen)
{
	emit(newData(buffer, buflen));
}


void MP3Stream::wroteStdin(KProcess */*proc*/)
{
	_free_to_write = true;
	timeout();
}


void MP3Stream::processExited(KProcess */*proc*/)
{
	// ACK! Shouldn't happen, unless one has a really old LAME version
	kdWarning(66666) << "LAME died unexpectedly. Disabling stream." << endl;
	stop();

	KMessageBox::sorry(0,
		i18n("The LAME encoder exited unexpectedly. Please make sure you are "
				"using a recent version of LAME, and check Floodcast's configuration "
				"in the Preferences dialog."));
	emit encoderDied();
}


void MP3Stream::stop()
{
	kdDebug(66666) << k_funcinfo << endl;
	// Clean up
	delete _lame;
	_lame = 0;

	delete _audioTimer;
	_audioTimer = 0;

	if (mRawStream)
	{
		napp->player()->engine()->visualizationStack()->remove(_mId);
		mRawStream->stop();

		delete mRawStream;
		mRawStream = 0;
	}
}


// Starts our decoder
void MP3Stream::start()
{
	kdDebug(66666) << k_funcinfo << "Creating artsd grabbing and mp3 encoding process" << endl;

	// create audio tap on server
	mRawStream = new Noatun::RawStream();
	*mRawStream = Arts::DynamicCast(
		napp->player()->engine()->server()->createObject("Noatun::RawStream"));

	if ((*mRawStream).isNull())
	{
		delete mRawStream;
		mRawStream = 0;
		kdDebug(66666) << k_funcinfo << "Couldn't create artsd rawstream!" << endl;
		return;
	}
	else
	{
		mRawStream->start();
		_mId = napp->player()->engine()->visualizationStack()->insertBottom(
			*mRawStream, "Icecast Audio Tap");
	}

	// Set up our RawStream polling timer
	_audioTimer = new QTimer(this, "audioTimer");
	connect(_audioTimer, SIGNAL(timeout()), this, SLOT(timeout()));

	_lame = new KProcess(this, "_lame");

	// Execute LAME
	*_lame << FloodcastConfig::lamePath();
	// Accept raw input
	*_lame << "-r";
	// Swap bytes (we produce litte endian, it wants big endian)
	*_lame << "-x";
	// either use joint-stereo or mono, depending on settings
	*_lame << "-m" << "j";
	//*_lame << (FloodcastConfig::mono() ? "m" : "j");
	if (FloodcastConfig::mono())
		*_lame << "-a";
	// Don't write a TAG, we're a stream
	*_lame << "-t";

	if (FloodcastConfig::bitrateType() == FloodcastConfig::Variable)
	{
		kdDebug(66666) << k_funcinfo << "Using VBR, quality = " << FloodcastConfig::vbrQuality() << endl;
		*_lame << "-V";
		*_lame << QString::number(FloodcastConfig::vbrQuality());
	}
	else if (FloodcastConfig::bitrateType() == FloodcastConfig::Average)
	{
		kdDebug(66666) << k_funcinfo << "Using ABR, bitrate = " << FloodcastConfig::bitrate() << endl;
		// Use average bitrate. Puts a bit more strain on an MP3 decoder's stream buffer, but
		// can greatly increase the (bandwidth):(audio quality) ratio
		*_lame << "--abr";
		*_lame << QString::number(FloodcastConfig::bitrate());
	}
	else
	{
		kdDebug(66666) << k_funcinfo << "Using CBR, bitrate = " << FloodcastConfig::bitrate() << endl;
		// Use constant bitrate
		*_lame << "-b";
		*_lame << QString::number(FloodcastConfig::bitrate());
	}

	// resample
	*_lame << "--resample";
	QString srate("44.1");
	switch (FloodcastConfig::samplerate())
	{
		case (FloodcastConfig::khz8):
			srate = "8";
			break;
		case (FloodcastConfig::khz11):
			srate = "11.025";
			break;
		case (FloodcastConfig::khz12):
			srate = "12";
			break;
		case (FloodcastConfig::khz16):
			srate = "16";
			break;
		case (FloodcastConfig::khz22):
			srate = "22.052";
			break;
		case (FloodcastConfig::khz24):
			srate = "24";
			break;
		case (FloodcastConfig::khz32):
			srate = "32";
			break;
		case (FloodcastConfig::khz44):
			srate = "44.1";
			break;
		case (FloodcastConfig::khz48):
			srate = "48";
			break;
	}
	*_lame << srate;
	// Does the user want high quality encoding? Unless you have a slow CPU,
	// this should always be on.
	if (FloodcastConfig::highQualityEncoding())
		*_lame << "-h";
	// If wanted, use error protection
	// This slightly lowers audio quality (needs 16bits per frame)
	if (FloodcastConfig::errorProtection())
		*_lame << "-p";
	// no annoying stdout messages, please
	*_lame << "--silent";
	// Get input from stdin, pipe output to stdout
	*_lame << "-" << "-";

	// Start our decoder
	if (!_lame->start(KProcess::NotifyOnExit, KProcess::All))
	{
		stop();
		emit encoderDied();
		KMessageBox::sorry(0,
			i18n("Floodcast was unable to start the LAME encoder. Please check " \
				"Floodcast's configuration in the Preferences dialog and make " \
				"sure LAME is installed."));
		return;
	}

	// Connect our signals
	connect(_lame, SIGNAL(receivedStdout(KProcess *, char *, int)),
		this, SLOT(receivedStdout(KProcess *, char *, int)));

	connect(_lame, SIGNAL(wroteStdin(KProcess *)),
		this, SLOT(wroteStdin(KProcess *)));

	connect(_lame, SIGNAL(processExited(KProcess *)), this,
		SLOT(processExited(KProcess *)));

	_audioTimer->start(DATACHECK_TIME, true);
	_free_to_write = true;
}

#include "mp3stream.moc"
