/*
    server.cpp  -  Floodcast server socket

    Copyright (c) 2002 by Ryan Cumming
    Copyright (c) 2004 by Stefan Gehn <metz AT gehn.net>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include "server.h"
#include "floodcastconfig.h"
#include "mp3stream.h"

#include <qregexp.h>
#include <qcstring.h>
#include <qtextstream.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kdebug.h>

//#define META_INTERVAL 8192
#define META_INTERVAL 32768


// ---------------------------------------------------------------------------


IcecastServerSocket::IcecastServerSocket(int port, QObject *parent,
	const char *name) : QServerSocket(port, 1, parent, name)
{
	if (!ok())
	{
		// Someone's already bound to our port, complain
		KMessageBox::sorry(0,
			i18n("Floodcast was unable to listen on port %1. Please make sure no" \
				" other application is using this port and try again.")
					.arg(QString::number(port)));
	}
}

void IcecastServerSocket::newConnection(int socket)
{
	// Let our parent know that we have a new connection
	emit(connection(socket));
}


// ---------------------------------------------------------------------------


IcecastClientSocket::IcecastClientSocket(int socket, IcecastServer *parent, const char *name)
	: QSocket(parent, name)
{
	mServer = parent;
	mUserAgent = i18n("Unknown");
	bSendMeta = false;
	mLastMetaData = 0;
	bStreaming = false;
	setSocket(socket);

	// Read client handles HTTP negotiation, they need to do this before they get an audio stream
	connect(this, SIGNAL(readyRead()), this, SLOT(slotRead()));
	// We want to know if this connection is closed, to we can reap it
	connect(this, SIGNAL(connectionClosed()), this, SLOT(slotConnectionClosed()));
}


void IcecastClientSocket::slotRead()
{
	if (!canReadLine())
		return;


	// Parse the HTTP request
	QStringList tokens = QStringList::split(QRegExp("[\\s]+"), readLine());
	//kdDebug(66666) << k_funcinfo << "tokens: " << tokens << endl;

	// Make sure it's a GET
	if (tokens[0] == "GET")
	{
		QTextStream os(this);
		// HTTP is all Latin1
		os.setEncoding(QTextStream::Latin1);

		if (!mServer->isActive())
		{
			// streaming disabled, don't accept incoming connections
			os << "ICY 401 Service Unavailable\r\n";
			flush();
			emit(clientDisconnected(this));
		}
		else if (mServer->connectedClients() >= mServer->maxClients())
		{
			// Server maxed out
			os << "ICY 503 Server Full\r\n";
			flush();
			emit(clientDisconnected(this));
		}
		else
		{
			while(canReadLine())
			{
				QStringList tokens = QStringList::split(QRegExp("[\\s]+"), readLine());
				//kdDebug(66666) << k_funcinfo << "tokens: " << tokens << endl;
				if (tokens[0] == "Icy-MetaData:1" || (tokens[0] == "Icy-MetaData:" && tokens[1] == "1"))
				{
					kdDebug(66666) << k_funcinfo <<
						"Client understands icy metadata" << endl;
					bSendMeta = true;
				}
				else if (tokens[0] == "User-Agent:")
				{
					mUserAgent = tokens[1];
					//kdDebug(66666) << k_funcinfo << "Useragent = " << mUserAgent << endl;
				}
			}

			// We understood the request, and we're an icecast server
			os << "ICY 200 OK\r\n";
			//os << "icy-notice1: Noatun Floodcast plugin\r\n";
			//os << "icy-notice2: Don't drink and drive\r\n";

			// Dump some stupid config settings
			QString title = FloodcastConfig::streamTitle();
			if (!title.isEmpty())
				os << "icy-name:" << title << "\r\n";

			QString genre = FloodcastConfig::streamGenre();
			if (!genre.isEmpty())
				os << "icy-genre:" << genre << "\r\n";

			QString url = FloodcastConfig::streamURL();
			if (!url.isEmpty())
				os << "icy-url:" << url << "\r\n";

			// And this is an MP3 audio stream
			os << "Content-Type:audio/mpeg\r\n";

			os << "icy-br:" << QString::number(FloodcastConfig::bitrate()) << "\r\n";

			// We don't do public streams
			// You should probably set up a full Icecast server for
			// anything more than a simple private audio stream
			os << "icy-pub:0\r\n";

			if (bSendMeta)
				os << "icy-metaint:" << QString::number(META_INTERVAL) << "\r\n";

			os << "\r\n";

			flush();

			// We don't care about any more read events, once the connection
			// is set up
			disconnect(this, SIGNAL(readyRead()), this, SLOT(slotRead()));

			// Let people that make be monitoring us (ie, the status window)
			// know about this connection
			emit(clientConnected(this));
			bStreaming = true;
		}
	}
	else
	{
		// Malformed request, dump the socket
		emit(clientDisconnected(this));
	}
}


void IcecastClientSocket::slotConnectionClosed()
{
	emit(clientDisconnected(this));
}


Q_LONG IcecastClientSocket::writeBlock(const char *data, Q_ULONG len)
{
	Q_LONG res = 0;

	if (!bStreaming) // Header sending, not yet sending out streaming data
	{
		res = QSocket::writeBlock(data, len);
	}
	else
	{
		if (bStreaming && bSendMeta && (len + mLastMetaData >= META_INTERVAL))
		{
			unsigned int i;
			unsigned int bytes_left = META_INTERVAL - mLastMetaData;
			unsigned int bytes_right = len - bytes_left;
			unsigned int pos = 0;

			//kdDebug(66666) << k_funcinfo << "mLastMetaData before sending metadata: " << mLastMetaData << endl;
			//kdDebug(66666) << k_funcinfo << "bytes_left = " << bytes_left  << endl;

			QCString metaString("StreamTitle='");
			metaString += mServer->title().latin1();
			metaString += "';StreamUrl='" + FloodcastConfig::streamURL() + "';";

			unsigned int metaSize = 1;
			if (metaString != mLastTitle)
				metaSize += (1 + int(metaString.size()/16)) * 16;

			QByteArray mBuf(len + metaSize);

			for(i=0; i < bytes_left; i++)
				mBuf[pos+i] = data[i];
			pos += bytes_left;

			if (metaSize == 1) // same title as last time, send null as length
			{
				//kdDebug(66666) << k_funcinfo << "sending empty title" << endl;
				mBuf[pos] = 0;
				pos++;
			}
			else
			{
				metaSize--; // metaSize includes the length byte, which we are going to add right now
				mBuf[pos] = (metaSize)/16;
				pos++;

				for (i=0; i < metaString.size(); i++)
					mBuf[pos+i] = metaString[i]; // add string bytes
				pos += metaString.size();
				metaSize -= metaString.size();

				for (i=0; i < metaSize; i++)
					mBuf[pos+i] = 0; // add padding null-bytes
				pos += metaSize;

				mLastTitle = metaString;
			}

			for(i=0; i < bytes_right; i++)
				mBuf[pos+i] = data[bytes_left+i];
			pos += bytes_right;

			//kdDebug(66666) << k_funcinfo << "initial writeBlock len: " << len << "; final length: " << mBuf.size() << endl;
			res = QSocket::writeBlock(mBuf.data(), mBuf.size());
			mLastMetaData = bytes_right;
			//kdDebug(66666) << k_funcinfo << "mLastMetaData is now: " << mLastMetaData << endl;
		}
		else
		{
			res = QSocket::writeBlock(data, len);
			mLastMetaData += len;
			//kdDebug(66666) << k_funcinfo << "sent " << len << " bytes; mLastMetaData now " << mLastMetaData << endl;
		}
	}
	return res;
}


const QString &IcecastClientSocket::userAgent()
{
	return mUserAgent;
}

// ---------------------------------------------------------------------------


IcecastServer::IcecastServer(QObject *parent, const char *name)
	: QObject(parent, name),
	mServerSocket(0),
	mStream(0),
	mActive(false)
{
	kdDebug(66666) << k_funcinfo << endl;

	mTitle = "";
	mClientSockets.setAutoDelete(true);
	mMaxClients = FloodcastConfig::serverMaxUsers();

	// This makes a new IcecastServerSocket
	setListenPort(FloodcastConfig::serverPort());
}


IcecastServer::~IcecastServer()
{
	kdDebug(66666) << k_funcinfo << endl;
}


void IcecastServer::addSocket(int socket)
{
	// Create the socket
	IcecastClientSocket* s = new IcecastClientSocket(socket, this);

	connect(s, SIGNAL(clientConnected(IcecastClientSocket *)),
		this, SLOT(slotClientConnected(IcecastClientSocket *)));
	connect(s, SIGNAL(clientDisconnected(IcecastClientSocket *)),
		this, SLOT(slotClientDisconnected(IcecastClientSocket *)));
}


void IcecastServer::slotClientConnected(IcecastClientSocket *clientSocket)
{
	//kdDebug(66666) << k_funcinfo << endl;

	// Will this be our first listener? Fire up the decoder
	if (mClientSockets.isEmpty())
	{
		if (mStream)
		{
			kdWarning(66666) << k_funcinfo <<
				"Creating a new mp3stream although we already have one!" << endl;
		}

		// Create the MP3 stream
		mStream = new MP3Stream(this, "mp3stream");
		// This signal is called when the encoder has produced new data
		connect(mStream, SIGNAL(newData(char *, int)),
			this, SLOT(transmitData(char *, int)));
		connect(mStream, SIGNAL(encoderDied()), this, SLOT(slotEncoderDied()));
	}

	// Add this socket to the list
	mClientSockets.append(clientSocket);
	emit gainedConnection(clientSocket);
}


void IcecastServer::slotClientDisconnected(IcecastClientSocket *clientSocket)
{
	//kdDebug(66666) << k_funcinfo << endl;
	removeSocket(clientSocket);
}


void IcecastServer::removeSocket(IcecastClientSocket *socket)
{
	//kdDebug(66666) << k_funcinfo << "removing socket" << endl;
	// Could we remove this from the socket list?
	if (!mClientSockets.remove(socket))
	{
		//kdDebug(66666) << k_funcinfo << "Deleting half-connected client-socket" << endl;
		// Nope, just delete it and hope for the best
		delete socket;

		// Note: gainedConnection() is only sent when
		// HTTP setup is complete, and a socket is
		// on the sockets list. Obviously, this socket
		// didn't make it that far, so don't emit
		// lostConnection()
	}
	else
	{
		// Yes, was this our last socket?
		if (mClientSockets.isEmpty())
		{
			//kdDebug(66666) << k_funcinfo << "last socket, closing down encoder" << endl;
			// Kill the encoder and aRts object
			delete mStream;
			mStream = 0;
		}

		// Let our monitors know about this connection's tragic death
		emit(lostConnection(socket));
	}
}


void IcecastServer::transmitData(char *buffer, int buflen)
{
	// We have new data to send
	for (IcecastClientSocket *s = mClientSockets.first(); s; s = mClientSockets.next())
	{
		// Is this socket still connected?
		if (s->state() == QSocket::Connected)
		{
			// Yes, send the data off
			if (s->writeBlock(buffer, buflen) == -1)
				removeSocket(s);
		}
		else
		{
			// No, clean up after it
			removeSocket(s);
		}
	}
}


void IcecastServer::restartEncoder()
{
	kdDebug(66666) << k_funcinfo << endl;
	// Restart our encoder object
	if (mStream)
	{
		delete mStream;
		mStream = new MP3Stream(this, "stream");
		connect(mStream, SIGNAL(newData(char *, int)),
			this, SLOT(transmitData(char *, int)));
		connect(mStream, SIGNAL(encoderDied()), this, SLOT(slotEncoderDied()));
	}
}


void IcecastServer::setListenPort(unsigned int port)
{
	if (mServerSocket)
	{
		// Don't recreate the socket if the listening port number didn't change
		if (mServerSocket->ok() && mServerSocket->port() == port)
			return;
	}

	// Destroy our old server socket, if any
	delete mServerSocket;
	// Make a new one
	mServerSocket = new IcecastServerSocket(port, this, "IcecastServersocket");
	// This is called once a new connection is made
	connect(mServerSocket, SIGNAL(connection(int)), this, SLOT(addSocket(int)));
}


int IcecastServer::maxClients()
{
	return mMaxClients;
}


void IcecastServer::setMaxClients(unsigned int max)
{
	mMaxClients = max;
}


void IcecastServer::stop()
{
	kdDebug(66666) << k_funcinfo << endl;
	if (!mClientSockets.isEmpty())
	{
		if (KMessageBox::questionYesNo(0,
			i18n("Do you really want to disconnect all listeners?"),
			i18n("Floodcast")) == KMessageBox::No)
		{
			emit(statusChange(true));
			return;
		}
	}
	shutdown();
}


void IcecastServer::start()
{
	kdDebug(66666) << k_funcinfo << endl;
	mActive = true;
	emit(statusChange(true));
}


void IcecastServer::slotEncoderDied()
{
	kdDebug(66666) << k_funcinfo << endl;
	shutdown();
}


void IcecastServer::shutdown()
{
	// We don't take new connections, sorry
	mActive = false;
	// And now kick anybody connected :)
	for (IcecastClientSocket *s = mClientSockets.first(); s; s = mClientSockets.next())
	{
		//kdDebug(66666) << k_funcinfo << "calling removeSocket()" << endl;
		removeSocket(s);
	}
	emit(statusChange(false));
}


void IcecastServer::setTitle(const QString &songtitle)
{
	mTitle = songtitle;
}

const QString & IcecastServer::title()
{
	return mTitle;
}

#include "server.moc"
