/*
    status.cpp  -  Floodcast standalone-server status-window

    Copyright (c) 2002 by Ryan Cumming
    Copyright (c) 2004 by Stefan Gehn <metz AT gehn.net>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include "server.h"
#include "status.h"
#include "plugin.h"
#include "floodcastconfig.h"

#include <qcheckbox.h>
#include <qfile.h>
#include <qxml.h>
#include <qvbox.h>
#include <qvgroupbox.h>
#include <qtimer.h>
#include <qwhatsthis.h>

#include <kdialogbase.h>
#include <kiconloader.h>
#include <kglobal.h>
#include <klocale.h>
#include <kio/netaccess.h>
#include <kmessagebox.h>
#include <klistview.h>
#include <kpopupmenu.h>
#include <kurl.h>

#include <kdebug.h>

#define POPUP_TITLE 1
#define COL_HOST 0
#define COL_TIME 1
#define COL_USERAGENT 2


class ShoutcastXMLStructure : public QXmlDefaultHandler
{
public:
	KListView *mListView;
	QString currentXMLItem, mCurrentUA, mCurrentHost, mCurrentConnectedSecs;
	bool mInsideListener;

	ShoutcastXMLStructure(KListView *lv)
	{
		mInsideListener = false;
		mListView = lv;
	}

	bool startElement(const QString &, const QString &, const QString &name, const QXmlAttributes &)
	{
		if (name.upper() == "LISTENER")
			mInsideListener = true;
		else if (mInsideListener)
			currentXMLItem = name.upper();
		return true;
	}

	bool endElement(const QString &, const QString &, const QString &name)
	{
		if (name.upper() == "LISTENER")
		{
			mInsideListener = false;
			//kdDebug(66666) << k_funcinfo << "LISTENER END" << endl;
			if (!mCurrentHost.isEmpty())
				new HostItem(mListView, mCurrentHost, mCurrentUA, mCurrentConnectedSecs.toInt());
			else
				kdDebug(66666) << k_funcinfo << "No HostAddress found in LISTENER object" << endl;

			currentXMLItem = QString::null;
			mCurrentHost = QString::null;
			mCurrentUA = QString::null;
			mCurrentConnectedSecs = QString::null;
		}
		else if (mInsideListener)
		{
			//kdDebug(66666) << k_funcinfo << "END LISTENER subobject " << name << endl;
			currentXMLItem = "";
		}
		return true;
	}

	bool characters(const QString &ch)
	{
		if (mInsideListener)
		{
			if (currentXMLItem == "USERAGENT")
				mCurrentUA = ch;
			else if (currentXMLItem == "HOSTNAME")
				mCurrentHost = ch;
			else if (currentXMLItem == "CONNECTTIME")
				mCurrentConnectedSecs = ch;
			/*else
			{
				kdDebug(66666) << k_funcinfo <<
					"currentXMLItem = " << currentXMLItem << ", ch = " << ch << endl;
			}*/
		}
		return true;
	}
};


// ----------------------------------------------------------------------------


HostItem::HostItem(KListView *parent, const QHostAddress &a, const QString &ua)
	: KListViewItem(parent)
{
	mAddress = a;
	mConnectTime = QDateTime::currentDateTime();

	setup(ua);

	mUpdateTimer = new QTimer(this, "mUpdateTimer");
	connect(mUpdateTimer, SIGNAL(timeout()), this, SLOT(updateTime()));
	mUpdateTimer->start(10 * 1000);
}


HostItem::HostItem(KListView *parent, const QString &a, const QString &ua, int connectedSecs)
	: KListViewItem(parent)
{
	mAddress.setAddress(a); // create QHostAddress from QString containing a readable IP
	mConnectTime = QDateTime::currentDateTime();
	mConnectTime = mConnectTime.addSecs(-connectedSecs); // subtract seconds this client is connected

	setup(ua);

	mUpdateTimer = 0L; // do not automatically update anything in Shoutcast mode!
}


void HostItem::setup(const QString &ua)
{
	kdDebug(66666) << k_funcinfo << "HostItem for " << mAddress.toString() << endl;
	mLookup = new QDns(mAddress);
	connect(mLookup, SIGNAL(resultsReady()), this, SLOT(slotLookupComplete()));
 	setPixmap(COL_HOST, SmallIcon("gear"));
	setText(COL_HOST, mAddress.toString());
	setText(COL_USERAGENT, ua);
	updateTime();
}


HostItem::~HostItem()
{
	delete mLookup;
	delete mUpdateTimer;
}


void HostItem::slotLookupComplete()
{
	QString name = mAddress.toString();

	QStringList hostNames = mLookup->hostNames();
	QStringList::Iterator it = hostNames.begin();
	while(it != hostNames.end())
	{
		if (!(*it).isEmpty())
		{
			name = (*it);
			break;
		}
		++it;
	}

	setText(COL_HOST, name);
	/*kdDebug(66666) << k_funcinfo <<
		"reverse lookup for HostItem IP complete: " << name << endl;*/
}


void HostItem::updateTime()
{
	unsigned int hours, minutes, seconds;
	QString time_string;

	seconds = mConnectTime.secsTo(QDateTime::currentDateTime());
	minutes = (seconds % 3600) / 60;
	hours = seconds / 3600;
	seconds -= hours * 3600;
	seconds -= minutes * 60;
	time_string.sprintf(i18n("hours:minutes:seconds", "%d:%.2d:%.2d"), hours, minutes, seconds);

	setText(COL_TIME, time_string);
}


// ----------------------------------------------------------------------------


FloodcastStatus::FloodcastStatus()
	: KDialogBase(0, "FloodcastStatus", false, i18n("Floodcast Status"),
		Close|User1, Close, false, KGuiItem(i18n("&Update"), "reload"))
{
	mServer = 0;

	connect(this, SIGNAL(closeClicked()), SLOT(hide()));

	QVBox *mWidget = makeVBoxMainWidget();

	mChkActive = new QCheckBox(i18n("&Use Floodcast"), mWidget, "chkActive");
	//connect(mChkActive, SIGNAL(toggled(bool)), this, SIGNAL(activeToggled()));

	QVGroupBox *groupBox = new QVGroupBox(i18n("Connected Users"), mWidget);
	groupBox->setMinimumWidth(380);

	mListView = new KListView(groupBox);
	mListView->addColumn(i18n("Hostname"));
	mListView->addColumn(i18n("Duration"));
	mListView->addColumn(i18n("Useragent"));
	mListView->setSelectionMode(QListView::Extended);
	mListView->setAllColumnsShowFocus(true);
	QWhatsThis::add(mListView,
		i18n("A list of currently connected users, and how long they have been connected "
			"in minutes. You can forcibly disconnect a user by right clicking on their "
			"list entry selecting \"Disconnect\"."));
	connect(mListView, SIGNAL(contextMenuRequested(QListViewItem *, const QPoint &, int)),
		this, SLOT(showContextMenu(QListViewItem *, const QPoint &)));

	mContextMenu = new KPopupMenu(this, "mContextMenu");
	mContextMenu->insertTitle("", POPUP_TITLE);
	mContextMenu->insertItem(
		KGlobal::iconLoader()->loadIconSet("connect_no", KIcon::Small),
		i18n("&Disconnect"), this, SLOT(disconnectClient()));
}


void FloodcastStatus::setServer(IcecastServer *server)
{
	if (server)
	{
		if (mServer)
			slotServerDestroyed();

		mServer = server;
		connect(mServer, SIGNAL(gainedConnection(IcecastClientSocket *)),
			this, SLOT(gainedConnection(IcecastClientSocket *)));
		connect(mServer, SIGNAL(lostConnection(IcecastClientSocket *)),
			this, SLOT(lostConnection(IcecastClientSocket *)));
		connect(mServer, SIGNAL(destroyed()), this, SLOT(slotServerDestroyed()));
	}
	else
	{
		slotServerDestroyed();
	}

	showButton(User1, (server == 0));
}


void FloodcastStatus::slotServerDestroyed()
{
	hide();
	disconnect(mServer, 0);
	mServer = 0;
}


void FloodcastStatus::gainedConnection(IcecastClientSocket *s)
{
	// Make a listview item
	HostItem *list_item = new HostItem(mListView, s->peerAddress(), s->userAgent());
	mEntryMap[s] = list_item;
}


void FloodcastStatus::lostConnection(IcecastClientSocket *s)
{
	if (mEntryMap.contains(s))
	{
		delete mEntryMap[s];
		mEntryMap.remove(s);
	}
}


void FloodcastStatus::disconnectClient()
{
	if (!mServer)
		return;

	QMap<IcecastClientSocket *, HostItem*>::iterator listview_it;
	QValueList<IcecastClientSocket*>::iterator sockets_it;
	QValueList<IcecastClientSocket*> dead_sockets;

	for (listview_it = mEntryMap.begin(); listview_it != mEntryMap.end(); ++listview_it)
	{
		if (listview_it.data()->isSelected())
		{
			dead_sockets.push_back(listview_it.key());
		}
	}

	for (sockets_it = dead_sockets.begin();sockets_it != dead_sockets.end(); ++sockets_it)
	{
		mServer->removeSocket(*sockets_it);
	}
}


void FloodcastStatus::showContextMenu(QListViewItem *it, const QPoint &pos)
{
	if (it && mServer)
	{
		mContextMenu->changeTitle(POPUP_TITLE, SmallIcon("gear"), it->text(COL_HOST));
		mContextMenu->exec(pos);
	}
}


void FloodcastStatus::slotUser1()
{
	QString dest;
	KURL url;
	url.setProtocol("http");
	url.setHost(FloodcastConfig::forwardHost());
	url.setPort(FloodcastConfig::forwardPort());
	url.setPath("/admin.cgi");
	url.setQuery("?pass=" + FloodcastConfig::forwardPassword() + "&mode=viewxml");

	if (KIO::NetAccess::download(url, dest, 0L))
	{
		QFile file(dest);
		if (file.open(IO_ReadOnly))
		{
			mListView->clear();
			QXmlInputSource source(&file);
			QXmlSimpleReader reader;

			ShoutcastXMLStructure parser(mListView);

			reader.setContentHandler(&parser);
			reader.parse(source);
		}
		KIO::NetAccess::removeTempFile(dest);
	}
	else
	{
		KMessageBox::sorry(0,
			i18n("Unable to fetch server-information right now"),
			i18n("Floodcast"));
	}
}


void FloodcastStatus::setActive(bool val)
{
	if (mChkActive->isChecked() == val)
		return;

	kdDebug(66666) << k_funcinfo << val << endl;
	mChkActive->blockSignals(true);
	mChkActive->setChecked(val);
	mChkActive->blockSignals(false);
}

#include "status.moc"
